#=======================================================================
#	$Id: QMap.pm,v 1.4 2010/03/26 08:45:43 pythontech Exp $
#	Qualified Name mapping
#=======================================================================
package PythonTech::QMap;
use strict;

#-----------------------------------------------------------------------
#	Create new qname map.
#	  qmap		Hash of: qprefix => expanded URI base e.g.
#			{
#			  c => 'http://meta/',
#			  codas => 'http://meta/',
#			  cs => 'http://meta/schema/',
#			  '' => 'http://example.com/foo#',
#			}
#			A mapping with an empty key means that relative
#			(or schemeless) URIs are interpreted as relative
#			to the given base URI
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {@_};
    bless $self, $class;
    die "$class: qmap not configured\n" unless $self->{'qmap'};

    $self->{'map'} = {%{$self->{'qmap'}}}; # Take a copy
    $self->{'rev'} = undef;	# Reverse map
    $self->{'regexp'} = undef;	# Regexp to match all expanded qprefixes
    $self->{'last'} = undef;	# Keep side effect for lastq method.
    return $self;
}

#-----------------------------------------------------------------------
#	Given a qname-abbreviated URI, expand to a full URI.
#	E.g.
#	  c:one		-> http://meta/one
#	  cs:two	-> http://meta/schema/two
#	  three		-> http://example.com/foo#three
#	  four dfltq=cs	-> http://meta/schema/four
#-----------------------------------------------------------------------
sub decode {
    my($self, $quri,$dfltq) = @_;
    return $quri if $quri =~ /^=/; # Convention for literal
    if (my($prefix,$rest) = $quri =~ /^([a-z][-\.\+a-z0-9]*):(.*)$/i) {
	if (defined(my $exp = $self->{'map'}->{$prefix})) {
	    $self->{'last'} = $prefix . ':';
	    return $exp . $rest;
	}
    } else {
	# No scheme - try given default qname, or mapping for ''
	$dfltq = '' unless defined $dfltq;
	if (defined(my $exp = $self->{'map'}->{$dfltq})) {
	    return $exp . $quri;
	}
    }
    return $quri;
}

#-----------------------------------------------------------------------
#	Given an arbitrary URI, contract it as mush as possible using
#	the defined qmap.  We want the longest matching expansion.
#	E.g.
#	  http://meta/one		-> c:one
#	  http://meta/schema/two	-> cs:two
#	  http://other/three		-> http://other/three
#	  http://example.com/foo#four	-> four
#-----------------------------------------------------------------------
sub encode {
    my($self, $uri) = @_;
#    print STDERR "_nq IN: $uri\n";
    if (! $self->{'rev'}) {
	# Build reverse qmap
	my $qmap = $self->{'map'};
	my %qrev;
	while (my($q,$e) = each %$qmap) {
	    $q .= ':' unless $q eq '';
	    # If more than one qprefix maps to same, choose shortest
	    $qrev{$e} = $q
		unless defined $qrev{$e} && length($qrev{$e}) < length($q);
	}
	$self->{'rev'} = \%qrev;
	# Build prefix-matching regexp.
	# N.B. we want to match longest first, so that e.g. with qmap
	#  a: -> http://foo/
	#  b: -> http://foo/bar/
	# then http://foo/bar/baz should be encoded as b:baz
	# rather than a:bar/baz .  It is sufficient to sort the
	# expansions into reverse order.
	# I don't know how clever Perl's regexp system is - will it
	# amalgamate alternatives to avoid backtracking? For example,
	# /^(abcdef|abc|abd)/ can be implemented as /^ab(c(def)?|d)/
	# (ignoring grouping).  If this turns out to be a hotspot,
	# we may consider making this optimisation manually.
	$self->{'regexp'} = ("^(" .
			     join("|",
				  map {quotemeta} 
				  sort {$b cmp $a}
				  values %$qmap) .
			     ")(.*)\$");
    }
    if (my($qexp,$rest) = $uri =~ $self->{'regexp'}) {
	my $prefix = $self->{'rev'}->{$qexp};
	$self->{'last'} = $prefix;
	$uri = $prefix . $rest;
    }
#    print STDERR "_nq OUT: $uri\n";
    return $uri;
}

sub lastq {
    my($self) = @_;
    return $self->{'last'};
}

1;
